#
# cs1robots.py
#
# Environment for steering a robot through a grid world
# for learning to program in Python
#
# Inspired by and using code from RUR-PLE by Andre Roberge
#
# 2010/01/14 Otfried Cheong
# 2010/02/23: Changed to refresh only once after every move
# 2010/02/24: Fixed refresh in world editing.
# 2010/08/12: Improved handling of beepers to speed up.
#
# On Linux, need packages python-tk, python-imaging-tk
#

import sys as _sys
import cs1graphics as _g
import easygui as _easygui
import re as _re
import time as _time

from cs1robots_images import _robot_images

# PIL isn't actually used in cs1robots, but is needed to use
# cs1graphics properly.  So we make sure it is there, as otherwise
# users get a confusing error message.
import PIL.Image as _Image
import PIL.ImageDraw as _ImageDraw
import PIL.ImageTk as _ImageTk

#_g._debug = 2

_scene = None
_world = None

_directions = [ (0, 1), (-1, 0), (0, -1), (1, 0) ]
_orient_dict = { 'E': 3, 'S': 2, 'W': 1, 'N': 0}

class RobotError(Exception):
  def __init__(self, str):
    Exception.__init__(self, str)

def pause(delay = 1.0):
  """Pause for delay seconds."""
  _time.sleep(delay)

# --------------------------------------------------------------------

class _Beeper(object):
  """One ore more beepers at a crossing in the world."""
  def __init__(self, radius, av, st, num = 1):
    self.av = av
    self.st = st
    self.num = num
    self.layer = _g.Layer()
    self.circle = _g.Circle(radius = radius)
    self.text = _g.Text("%d" % num, 11, _g.Point(0, 0))
    self.layer.add(self.circle)
    self.layer.add(self.text)
    self.circle.setFillColor("yellow")
    self.circle.setBorderColor("orange")
    self.circle.setDepth(10)
    self.text.setDepth(5)
    self.layer.setDepth(5)

  def set_number(self, num):
    self.num = num
    self.text.setMessage("%d" % num)
    
class _World(object):
  """This class defines the world's logic."""

  def __init__(self, avenues = 10, streets = 10, walls = [], beepers = {}):
    self.av = avenues
    self.st = streets
    self.num_cols = 2*avenues + 1
    self.num_rows = 2*streets + 1
    self.walls = walls
    for (col, row) in self.walls:
      if not (col+row) % 2:
        raise RuntimeError("Wall in impossible position (%d, %d)." % (col,row))
    self.beepers = beepers
    self.borders = []
    self.beeper_icons = {}
    self.set_borders()

  def cr2xy(self, col, row):
    x = self.left + self.ts * col
    y = self.bottom - self.ts * row
    return x, y

  def set_borders(self):
    """The world is surrounded by a continuous wall.  This function
       sets the corresponding "wall" or "border" based on the world's
       dimensions."""
    for col in range(1, self.num_cols-1, 2):
      if (col, 0) not in self.borders:
        self.borders.append( (col, 0) )
      if (col, self.num_rows) not in self.borders:
        self.borders.append( (col, self.num_rows-1) )
      for row in range(1, self.num_rows-1, 2):
        if (0, row) not in self.borders:
          self.borders.append( (0, row) )
        if (self.num_cols, row) not in self.borders:
          self.borders.append( (self.num_cols-1, row) )

  def toggle_wall(self, col, row):
    """This function is intended for adding or removing a
       wall from a GUI world editor."""
    if (col+row) % 2 :  # safety check
      if (col, row) in self.walls: # toggle value
        self.walls.remove((col, row))
      else:
        self.walls.append((col, row))
    else:
      raise RuntimeError("Wall in impossible position (%d, %d)." % (col,row))

  def is_clear(self, col, row):
    """Returns True if there is no wall or border here."""
    return not ((col, row) in self.walls or
                (col, row) in self.borders)
  
  def add_beeper(self, av, st):
    """Add a single beeper."""
    assert self.layer
    if (av, st) in self.beepers:
      self.beepers[(av, st)] += 1
      bp = self.beeper_icons[(av,st)]
      bp.set_number(self.beepers[(av, st)])
    else:
      self.beepers[(av, st)] = 1
      self._create_beeper(av, st)

  def remove_beeper(self, av, st):
    """Remove a beeper (does nothing if no beeper here)."""
    assert self.layer
    if (av, st) in self.beepers:
      self.beepers[(av, st)] -= 1
      if self.beepers[(av, st)] == 0:
        del self.beepers[(av, st)]
        self.layer.remove(self.beeper_icons[(av,st)].layer)
        del self.beeper_icons[(av,st)]
      else:
        bp = self.beeper_icons[(av,st)]
        bp.set_number(self.beepers[(av, st)])

  def _create_beeper(self, av, st):
    num = self.beepers[(av, st)]
    bp = _Beeper(0.6 * self.ts, av, st, num)
    x, y = self.cr2xy(2 * av - 1, 2 * st - 1)
    bp.layer.moveTo(x, y)
    self.beeper_icons[(av, st)] = bp
    self.layer.add(bp.layer)

  def create_layer(self):
    # compute tilesize and border coordinates
    w, h = _scene.getWidth(), _scene.getHeight()
    tsx =  w / (self.num_cols + 2)
    tsy =  h / (self.num_rows + 2)
    self.ts = min(tsx, tsy)
    self.left = 2 * self.ts
    self.right = self.left + 2 * self.ts * self.av
    self.bottom = h - 2 * self.ts
    self.top = self.bottom - 2 * self.ts * self.st
    self.layer = _g.Layer()
    # Create avenues
    for av in range(self.av):
      x = self.left + self.ts * (2 * av + 1) 
      l = _g.Path([_g.Point(x, self.top), _g.Point(x, self.bottom)])
      l.setBorderColor("light gray")
      self.layer.add(l)
      self.layer.add(_g.Text("%d" % (av + 1), 10, 
                             _g.Point(x, self.bottom + self.ts)))
    # Create streets
    for st in range(self.st):
      y = self.bottom - self.ts * (2 * st + 1) 
      l = _g.Path([_g.Point(self.left, y), _g.Point(self.right, y)])
      l.setBorderColor("light gray")
      self.layer.add(l)
      self.layer.add(_g.Text("%d" % (st + 1), 10, 
                             _g.Point(self.left - self.ts, y)))
    # Create border
    border = _g.Polygon(_g.Point(self.left, self.bottom),
                        _g.Point(self.right, self.bottom),
                        _g.Point(self.right, self.top),
                        _g.Point(self.left, self.top))
    border.setBorderWidth(5)
    border.setBorderColor("dark red")
    border.setDepth(10)
    self.layer.add(border)
    # Create walls
    for (col, row) in self.walls:
      if col % 2 == 0:  # vertical wall
        x1, y1 = self.cr2xy(col, row - 1)
        x2, y2 = self.cr2xy(col, row + 1)
      else: # horizontal wall
        x1, y1 = self.cr2xy(col - 1, row)
        x2, y2 = self.cr2xy(col + 1, row)
      w = _g.Path([_g.Point(x1,y1), _g.Point(x2,y2)])
      w.setBorderWidth(5)
      w.setBorderColor("dark red")
      w.setDepth(10)
      self.layer.add(w)
    # Create beepers
    for (av, st) in self.beepers:
      self._create_beeper(av, st)
    # Layer finished
    _scene.add(self.layer)
    
  def update_layer(self):
    _scene.remove(self.layer)
    self.beeper_icons = {}
    self.create_layer()

  def save(self, out):
    av_string = "avenues = " + str(self.av) + '\n'
    st_string = "streets = " + str(self.st) + '\n'
    if len(self.walls) > 0:
      wall_string = "walls = [\n"
      for item in self.walls:
        wall_string += ("    " + str(item) + ', \n')
      wall_string = wall_string[:-3] + '\n]\n'
    else:
      wall_string = "walls = []\n"
    if len(self.beepers) > 0:
      beeper_string = "beepers = {\n"
      for key in self.beepers:
        beeper_string += ("    " + str(key) + ': ' +
                          str(self.beepers[key]) + ', \n')
      beeper_string = beeper_string[:-3] + '\n}\n'
    else:
      beeper_string = "beepers = {}\n"
    out.write(av_string + st_string + wall_string + beeper_string)
    
# --------------------------------------------------------------------

def create_world(avenues = 10, streets = 10):
  """Create an empty robot world."""
  global _scene, _world
  if _scene:
    raise RuntimeError("A robot world already exists!")
  _scene = _g.Canvas()
  _scene.setWidth(50 * avenues)
  _scene.setHeight(50 * streets)
  _scene.setTitle("Robot World")
  _world = _World(avenues, streets)
  _world.create_layer()
  _scene.setAutoRefresh(False)

def _check_world(contents):
  # safety check
  safe = contents[:]
    # only allow known keywords
  keywords = ["avenues", "streets", "walls", "beepers", "robot",
              "'s'", "'S'", '"s"', '"S"',
              "'e'", "'E'", '"e"', '"E"',
              "'w'", "'W'", '"w"', '"W"',
              "'n'", "'N'", '"n"', '"N"', ]
  for word in keywords:
    safe = safe.replace(word, '')
  safe = list(safe)
  for char in safe:
    if char.isalpha():
      raise ValueError("Invalid word or character in world file")
    
def load_world(filename = None):
  """Loads a robot world from filename.
  Opens file-chooser if no filename is given."""
  global _scene, _world
  if _scene:
    raise RuntimeError("A robot world already exists!")
  if not filename:
    filename = _easygui.fileopenbox("Select a Robot world", 
                                    "Robot World", '*', [ "*.wld" ])
    if not filename: 
      raise RuntimeError("No world file selected.")
  txt = open(filename, 'rb').read()
  txt = _re.sub('\r\n', '\n', txt) # Windows
  txt = _re.sub('\r', '\n', txt)  # Mac
  _check_world(txt)
  wd = {}
  # extracts avenues, streets, walls and beepers
  try:
    exec (txt in wd)
    w = _World(wd['avenues'], wd['streets'], wd['walls'], wd['beepers'])
  except:
    raise ValueError("Error interpreting world file.")
  _world = w
  _scene = _g.Canvas()
  _scene.setWidth(50 * w.av)
  _scene.setHeight(50 * w.st)
  i = filename.rfind("/")
  if i >= 0: filename = filename[i+1:]
  _scene.setTitle("Robot World: " + filename)
  _world.create_layer()
  _scene.setAutoRefresh(False)

def save_world(filename = None):
  """Save a robot world to filename.
  Opens file-chooser if no filename is given."""
  if not filename:
    filename = _easygui.filesavebox("Select a Robot world", 
                                    "Robot World", '*', [ "*.wld" ])
    if not filename: 
      raise RuntimeError("No world file selected.")
  out = open(filename, "w")
  _world.save(out)
  out.close()

def edit_world():
  """Edit the current robot world.
  You can add and remove walls by clicking at the wall position.
  Add a beeper by clicking with the left button at an intersection.
  Remove a beeper with the right mouse button."""
  while True:
    e = _scene.wait()
    d = e.getDescription()
    if d == "canvas close":
      _sys.exit(1)
    if d == "keyboard" and e.getKey() in [ '\r', '\n' ]:
      return
    if d == "mouse click":
      x = int(e.getMouseLocation().getX())
      y = int(e.getMouseLocation().getY())
      print ("Mouse button", e.getButton(), "at", (x, y))
      col = (x - _world.left + _world.ts / 2) / _world.ts
      row = (_world.bottom - y + _world.ts / 2) / _world.ts
      if (col % 2) == 1 and (row % 2) == 1:
        print ("corner")
        # corner
        av = (col + 1) / 2
        st = (row + 1) / 2
        if av < 1 or av > _world.av or st < 1 or st > _world.st:
          continue
        if e.getButton() == 1:
          print ("add beeper")
          _world.add_beeper(av, st)
          _scene.refresh()
        elif e.getButton() == 3:
          print ("remove beeper")
          _world.remove_beeper(av, st)
          _scene.refresh()
      elif ((col + row) % 2) == 1:
        # wall position
        print ("wall position")
        if (col < 1 or col >= _world.num_cols - 1 or
            row < 1 or row >= _world.num_rows - 1):
          continue
        _world.toggle_wall(col, row)
        _world.update_layer()
        _scene.refresh()
  
# --------------------------------------------------------------------

class Robot(object):

  def __init__(self, color = "gray", orientation = 'E', beepers = 0,
               avenue = 1, street = 1):
    """Create a new robot."""
    if not color in _robot_images:
      raise TypeError('color must be a color name')
    self._image = [ None, None, None, None ]
    for i in range(4):
      self._image[i] = _g.Image("base64:" + _robot_images[color][i])
      self._image[i].moveTo(-100, -100)
      self._image[i].setDepth(0)
      _scene.add(self._image[i])
    self._dir = _orient_dict[orientation]
    self._x = avenue
    self._y = street
    self._beeper_bag = beepers
    self._trace = None
    self._delay = 0
    self._steps = 0
    self._update_pos()
    _scene.refresh()

  def _update_pos(self):
    x, y  = _world.cr2xy(2 * self._x - 1, 2 * self._y - 1)
    self._image[self._dir].moveTo(x, y)

  def _trace_pos(self):
    x, y  = _world.cr2xy(2 * self._x - 1, 2 * self._y - 1)
    xr, yr =  _directions[(self._dir - 1) % 4]
    xb, yb =  _directions[(self._dir - 2) % 4]
    return x + 5 * (xr + xb), y - 5 * (yr + yb)

  def _update_trace(self):
    if self._trace:
      x, y = self._trace_pos()
      self._trace.addPoint(_g.Point(x, y))

  def _refresh(self):
    _scene.refresh()
      
  def __del__(self):
    if _scene:
      for i in range(4):
        _scene.remove(self._image[i])

# --------------------------------------------------------------------

  def set_trace(self, color = None):
    """Without color argument, turn off tracing.
With color argument, start a new trace in that color."""
    if not color:
      if self._trace:
        _scene.remove(self._trace)
      self._trace = None
    else:
      x, y  = self._trace_pos()
      self._trace = _g.Path([_g.Point(x, y)])
      self._trace.setBorderColor(color)
      _scene.add(self._trace)

  def set_pause(self, delay = 0):
    """Set a pause to be made after each move."""
    self._delay = delay

# --------------------------------------------------------------------

  def get_pos(self):
    """Returns current robot position."""
    return self._x, self._y

  def turn_left(self):
    """Rotate left by 90 degrees."""
    self._image[self._dir].moveTo(-100, -100)
    self._dir = (self._dir + 1) % 4
    self._update_pos()
    self._update_trace()
    self._refresh()

  def move(self):
    """Move one street/avenue in direction where robot is facing."""
    if self.front_is_clear():
      xx, yy = _directions[self._dir]
      self._x += xx
      self._y += yy
    else:
      raise RobotError("That move really hurt!\n Please, make sure that " + 
                       "there is no wall in front of me!""")
    self._update_pos()
    self._update_trace()
    self._refresh()
    if self._delay > 0:
      _time.sleep(self._delay)
    
  def front_is_clear(self):
    """Returns True if no wall or border in front of robot."""
    col = 2 * self._x - 1
    row = 2 * self._y - 1
    xx, yy = _directions[self._dir]
    return _world.is_clear(col + xx, row + yy)

  def left_is_clear(self):
    """Returns True if no walls or borders are to the immediate left
    of the robot."""
    col = 2 * self._x - 1
    row = 2 * self._y - 1
    xx, yy = _directions[(self._dir + 1) % 4]
    return _world.is_clear(col + xx, row + yy)

  def right_is_clear(self):
    """Returns True if no walls or borders are to the immediate right
    of the robot."""
    col = 2 * self._x - 1
    row = 2 * self._y - 1
    xx, yy = _directions[(self._dir - 1) % 4]
    return _world.is_clear(col + xx, row + yy)

  def facing_north(self):
    """Returns True if Robot is facing north."""
    return (self._dir == 0)

  def carries_beepers(self):
    """Returns True if some beepers are left in Robot's bag."""
    return (self._beeper_bag > 0)

  def on_beeper(self):
    """Returns True if beepers are present at current robot position."""
    return ((self._x, self._y) in _world.beepers)

  def pick_beeper(self):
    """Robot picks one beeper up at current location."""
    if self.on_beeper():
      _world.remove_beeper(self._x, self._y)
      self._refresh()
      self._beeper_bag += 1
    else:
      raise RobotError("I must be on a beeper to pick it up.")

  def drop_beeper(self):
    """Robot drops one beeper down at current location."""
    if self.carries_beepers():
      self._beeper_bag -= 1
      _world.add_beeper(self._x, self._y)
      self._refresh()
    else:
      raise RobotError("I am not carrying any beepers.")

# --------------------------------------------------------------------

if __name__ == "__main__":
  if len(_sys.argv) > 1:
    load_world(_sys.argv[1])
  else:
    create_world(24, 14)
  edit_world()
  r = Robot("yellow")
  r.set_trace("blue")
  while True:
    try:
      e = _scene.wait()
      d = e.getDescription()
      if d == "canvas close":
        break
      if d == "keyboard":
        k = e.getKey()
        if k == "q":
          _scene.close()
          break
        elif k == " ":
          try:
            r.move()
          except:
            _easygui.msgbox("OUCH!")
        elif k == "t":
          r.turn_left()
        elif k == "w":
          load_world()
        elif k == "p":
          r.pick_beeper()
        elif k == "d":
          r.drop_beeper()
    except:
      _easygui.exceptionbox()

# --------------------------------------------------------------------



#
# cs1media.py
#
# Environment for manipulating Images 
#
# 2010/03/24 Otfried Cheong
# 2010/09/02: Convert image to RGB on loading
#
# Inspired and using some code from picture.py by Mark Guzdial.
#
# On Linux, need packages python-tk, python-imaging-tk
#

import sys as _sys
import PIL.Image as _Image
import PIL.ImageTk as _ImageTk
import easygui as _easygui
import tkinter.colorchooser as _tkColorChooser
import tkinter.font as _tkFont
import tkinter as _Tk

# --------------------------------------------------------------------

class Picture(object):
  """A digital image."""

  def __init__(self, surf):
    """Create a Picture from an Image object."""
    self._title = ""
    self._reset(surf)

  def _reset(self, surf):
    self._surf = surf
    self._pixels = surf.load()

  def size(self):
    """Return size of the image as a tuple (width, height)."""
    return self._surf.size

  def show1(self):
    """Display the image."""
    self._surf.show()

  def show(self):
    """Display the image and wait until user closes the image window."""
    tool = PictureTool(self)
    tool.run_tool()

  def set_pixels(self, color = (0,0,0)):
    """Set all pixels of the image to the given color."""
    surf = _Image.new("RGB", self._surf.size, color)
    self._reset(surf)

  def set_title(self, title):
    """Set title of image."""
    self._title = title

  def title(self):
    """Return title of image."""
    return self._title

  def get(self, x, y):
    """Return pixel at x, y."""
    return self._pixels[x, y]

  def set(self, x, y, color):
    """Set pixel at x, y to color."""
    self._pixels[x, y] = color

  def save_as(self, filename = None):
    """Save image as filename.
    If no filename is given, open file-chooser."""
    if not filename:
      filename = _easygui.filesavebox("Save image as", _sys.argv[0], 
                                      "unnamed.png",
                                      [ [ "*.jpg", "*.png", "*.bmp",
                                          "Image files" ] ])
      if not filename: 
        raise RuntimeError("No file name provided for saving.")
    self._surf.save(filename)

# --------------------------------------------------------------------

def create_picture(width, height, color = (0,0,0)):
  """Create an image of size width x height, and fill with color."""
  if width < 0 or height < 0:
    raise ValueError("Invalid image dimensions: " + str(width) + ", " 
                     + str(height))
  p = Picture(_Image.new("RGB", (width, height), color))
  return p

def load_picture(filename = None):
  """Create an image by loading file filename.
  Opens file-chooser if no file name given."""
  if not filename:
    filename = _easygui.fileopenbox("Select an image", 
                                    _sys.argv[0], '*', 
                                    [ [ "*.jpg", "*.png", "*.bmp", "*.gif",
                                        "Image files" ] ])
    if not filename: 
      raise RuntimeError("No image file selected.")
  img = _Image.open(filename)
  if img.mode != "RGB":
    img = img.convert("RGB")
  p = Picture(img)
  p.set_title(filename)
  return p

def choose_color():
  color = _tkColorChooser.askcolor()
  new_color = (color[0][0], color[0][1], color[0][2])
  return new_color

# --------------------------------------------------------------------

##
## Color Constants
##

class Color(object):
  """Definitions for many beautiful colors."""

  aliceblue = (240, 248, 255)
  antiquewhite = (250, 235, 215)
  aqua = (0, 255, 255)
  aquamarine = (127, 255, 212)
  azure = (240, 255, 255)
  beige = (245, 245, 220)
  bisque = (255, 228, 196)
  black = (0, 0, 0)
  blanchedalmond = (255, 235, 205)
  blue = (0, 0, 255)
  blueviolet = (138, 43, 226)
  brown = (165, 42, 42)
  burlywood = (222, 184, 135)
  cadetblue = (95, 158, 160)
  chartreuse = (127, 255, 0)
  chocolate = (210, 105, 30)
  coral = (255, 127, 80)
  cornflowerblue = (100, 149, 237)
  cornsilk = (255, 248, 220)
  crimson = (220, 20, 60)
  cyan = (0, 255, 255)
  darkblue = (0, 0, 139)
  darkcyan = (0, 139, 139)
  darkgoldenrod = (184, 134, 11)
  darkgray = (169, 169, 169)
  darkgreen = (0, 100, 0)
  darkkhaki = (189, 183, 107)
  darkmagenta = (139, 0, 139)
  darkolivegreen = (85, 107, 47)
  darkorange = (255, 140, 0)
  darkorchid = (153, 50, 204)
  darkred = (139, 0, 0)
  darksalmon = (233, 150, 122)
  darkseagreen = (143, 188, 143)
  darkslateblue = (72, 61, 139)
  darkslategray = (47, 79, 79)
  darkturquoise = (0, 206, 209)
  darkviolet = (148, 0, 211)
  deeppink = (255, 20, 147)
  deepskyblue = (0, 191, 255)
  dimgray = (105, 105, 105)
  dodgerblue = (30, 144, 255)
  firebrick = (178, 34, 34)
  floralwhite = (255, 250, 240)
  forestgreen = (34, 139, 34)
  fuchsia = (255, 0, 255)
  gainsboro = (220, 220, 220)
  ghostwhite = (248, 248, 255)
  gold = (255, 215, 0)
  goldenrod = (218, 165, 32)
  gray = (128, 128, 128)
  green = (0, 128, 0)
  greenyellow = (173, 255, 47)
  honeydew = (240, 255, 240)
  hotpink = (255, 105, 180)
  indianred = (205, 92, 92)
  indigo = (75, 0, 130)
  ivory = (255, 255, 240)
  khaki = (240, 230, 140)
  lavender = (230, 230, 250)
  lavenderblush = (255, 240, 245)
  lawngreen = (124, 252, 0)
  lemonchiffon = (255, 250, 205)
  lightblue = (173, 216, 230)
  lightcoral = (240, 128, 128)
  lightcyan = (224, 255, 255)
  lightgoldenrodyellow = (250, 250, 210)
  lightgreen = (144, 238, 144)
  lightgrey = (211, 211, 211)
  lightpink = (255, 182, 193)
  lightsalmon = (255, 160, 122)
  lightseagreen = (32, 178, 170)
  lightskyblue = (135, 206, 250)
  lightslategray = (119, 136, 153)
  lightsteelblue = (176, 196, 222)
  lightyellow = (255, 255, 224)
  lime = (0, 255, 0)
  limegreen = (50, 205, 50)
  linen = (250, 240, 230)
  magenta = (255, 0, 255)
  maroon = (128, 0, 0)
  mediumaquamarine = (102, 205, 170)
  mediumblue = (0, 0, 205)
  mediumorchid = (186, 85, 211)
  mediumpurple = (147, 112, 219)
  mediumseagreen = (60, 179, 113)
  mediumslateblue = (123, 104, 238)
  mediumspringgreen = (0, 250, 154)
  mediumturquoise = (72, 209, 204)
  mediumvioletred = (199, 21, 133)
  midnightblue = (25, 25, 112)
  mintcream = (245, 255, 250)
  mistyrose = (255, 228, 225)
  moccasin = (255, 228, 181)
  navajowhite = (255, 222, 173)
  navy = (0, 0, 128)
  oldlace = (253, 245, 230)
  olive = (128, 128, 0)
  olivedrab = (107, 142, 35)
  orange = (255, 165, 0)
  orangered = (255, 69, 0)
  orchid = (218, 112, 214)
  palegoldenrod = (238, 232, 170)
  palegreen = (152, 251, 152)
  paleturquoise = (175, 238, 238)
  palevioletred = (219, 112, 147)
  papayawhip = (255, 239, 213)
  peachpuff = (255, 218, 185)
  peru = (205, 133, 63)
  pink = (255, 192, 203)
  plum = (221, 160, 221)
  powderblue = (176, 224, 230)
  purple = (128, 0, 128)
  red = (255, 0, 0)
  rosybrown = (188, 143, 143)
  royalblue = (65, 105, 225)
  saddlebrown = (139, 69, 19)
  salmon = (250, 128, 114)
  sandybrown = (244, 164, 96)
  seagreen = (46, 139, 87)
  seashell = (255, 245, 238)
  sienna = (160, 82, 45)
  silver = (192, 192, 192)
  skyblue = (135, 206, 235)
  slateblue = (106, 90, 205)
  slategray = (112, 128, 144)
  snow = (255, 250, 250)
  springgreen = (0, 255, 127)
  steelblue = (70, 130, 180)
  tan = (210, 180, 140)
  teal = (0, 128, 128)
  thistle = (216, 191, 216)
  tomato = (255, 99, 71)
  turquoise = (64, 224, 208)
  violet = (238, 130, 238)
  wheat = (245, 222, 179)
  white = (255, 255, 255)
  whitesmoke = (245, 245, 245)
  yellow = (255, 255, 0)
  yellowgreen = (154, 205, 50)
  
# --------------------------------------------------------------------

class PictureTool:

  def __init__(self, pict):
    self.pict = pict
    
  def run_tool(self):
    self.root = _Tk.Tk()
    
    self.top = _Tk.Menu(self.root, bd=2)
    self.root.config(menu=self.top)
    
    self.zoom = _Tk.Menu(self.top, tearoff=0)
    self.zoom.add_command(label='25%', command=lambda : self.zoomf(0.25),
                          underline=0)
    self.zoom.add_command(label='50%', command=lambda : self.zoomf(0.5),
                          underline=0)
    self.zoom.add_command(label='75%', command=lambda : self.zoomf(0.75),
                          underline=0)
    self.zoom.add_command(label='100%', command=lambda : self.zoomf(1.0),
                          underline=0)
    self.zoom.add_command(label='150%', command=lambda : self.zoomf(1.5),
                          underline=0)
    self.zoom.add_command(label='200%', command=lambda : self.zoomf(2.0),
                          underline=0)
    self.zoom.add_command(label='400%', command=lambda : self.zoomf(4.0),
                          underline=0)
    self.zoom.add_command(label='800%', command=lambda : self.zoomf(8.0),
                          underline=0)
    
    self.top.add_cascade(label='Zoom', menu=self.zoom, underline=0)
    
    # create a frame and pack it
    self.frame1 = _Tk.Frame(self.root)
    self.frame1.pack(side=_Tk.BOTTOM, fill=_Tk.X)
    
    self.root.im = self.pict._surf
    self.root.zoomMult = 1.0
    
    self.root.photo1 = _ImageTk.PhotoImage(image=self.root.im)
    
    self.root.title(self.pict.title())
    
    # Canvas for the Image, with scroll bars
    
    self.canvas1 = _Tk.Canvas(self.frame1, width=self.root.photo1.width() -
                              1, height=self.root.photo1.height() - 1,
                              cursor="crosshair", borderwidth=0)
    self.root.vbar = _Tk.Scrollbar(self.frame1)
    self.root.hbar = _Tk.Scrollbar(self.frame1, orient='horizontal')
    self.root.vbar.pack(side=_Tk.RIGHT, fill=_Tk.Y)
    self.root.hbar.pack(side=_Tk.BOTTOM, fill=_Tk.X)
    
    self.canvas1.pack(side=_Tk.BOTTOM, padx=0, pady=0, anchor=_Tk.NW, 
                      fill=_Tk.BOTH, expand=_Tk.YES)

    # call on scroll move
    self.root.vbar.config(command=self.canvas1.yview)  
    self.root.hbar.config(command=self.canvas1.xview)
    # call on canvas move
    self.canvas1.config(yscrollcommand=self.root.vbar.set)  
    self.canvas1.config(xscrollcommand=self.root.hbar.set)
    self.draw_image(self.root.im)
    self.canvas1.bind('<Button-1>', self.canvClick)
    
    self.v = _Tk.StringVar()
    self.v.set("R:      G:      B:     ")
    self.xy = _Tk.StringVar()
    self.xy.set("X:      Y:      ")
    row = _Tk.Frame(self.root)
    font = _tkFont.Font(size=10)
    xyLabel = _Tk.Label(row, textvariable=self.xy, font=font)
    colorLabel = _Tk.Label(row, textvariable=self.v, font=font)
    self.canvas2 = _Tk.Canvas(row, width=35, bd=2, relief=_Tk.RIDGE, height=30)
    xyLabel.pack(side=_Tk.LEFT)
    colorLabel.pack(side=_Tk.LEFT, padx=100, pady=1)
    self.canvas2.pack(side=_Tk.LEFT, padx=2, pady=1)
    row.pack(side=_Tk.TOP, fill=_Tk.X)  # pack row on top

    # start the event loop
    self.root.mainloop()
      
  def zoomf(self, factor):
    # zoom in or out
    self.root.zoomMult = factor
    (wide, high) = self.root.im.size
    new = self.root.im.resize((int(wide * factor), int(high * factor)))
    self.draw_image(new)

  def draw_image(self, imgpil):
    self.root.photo1 = _ImageTk.PhotoImage(image=imgpil)  # not file=imgpath
    (scrwide, scrhigh) = self.root.maxsize()  # wm screen size x,y
    scrhigh -= 200  # leave room for top display/button at max photo size
    imgwide = self.root.photo1.width()  # size in pixels
    imghigh = self.root.photo1.height()  # same as imgpil.size
    
    fullsize = (0, 0, imgwide, imghigh)  # scrollable
    viewwide = min(imgwide, scrwide)  # viewable
    viewhigh = min(imghigh, scrhigh)
    
    self.canvas1.delete('all')  # clear prior photo
    self.canvas1.config(height=viewhigh, width=viewwide)  # viewable window size
    self.canvas1.config(scrollregion=fullsize)  # scrollable area size
    
    self.root.img = self.canvas1.create_image(0, 0, image=self.root.photo1,
                                              anchor=_Tk.NW)
    
    if imgwide <= scrwide and imghigh <= scrhigh:  # too big for display?
      self.root.state('normal')  # no: win size per img
    elif (_sys.platform)[:3] == 'win':
      # do windows fullscreen
      self.root.state('zoomed')  # others use geometry( )

  def canvClick(self, event):
    try:
      x = int(self.canvas1.canvasx(event.x) / self.root.zoomMult)
      y = int(self.canvas1.canvasy(event.y) / self.root.zoomMult)
      w, h = self.root.im.size
      if 0 <= x < w and 0 <= y < h:
        tk_rgb = "#%02x%02x%02x" % self.root.im.getpixel((x, y))
        self.canvas2.config(bg=tk_rgb)
        rgb = "R: %d; G: %d; B: %d;" % self.root.im.getpixel((x, y))
        self.v.set(rgb)
        xy = "X: %d; Y: %d;" % (x, y)
        self.xy.set(xy)
      else:
        rgb = "X,Y Out of Range"
        self.v.set(rgb)
    except ValueError:
      pass
    
def picture_tool(filename = None):
  """Allows you to find information about digital images.

  The PictureTool's Toolbar:

  Once you have opened an image, you can view information about its
  individual pixels by looking at the toolbar. To select a pixel drag
  (click and hold down) the mouse to the position you want and then
  release it to hold that position's information in the toolbar.

  The following information in the toolbar changes to reflect the
  properties of the pixel you selected:

  X = the x coordinate of the pixel (starting with 0, counting from the left) 
  Y = the y coordinate of the pixel (starting with 0, counting from the top)
  R = the Red value of the pixel (0 to 255)
  G = the Green value of the pixel (0 to 255)
  B = the Blue value of the pixel (0 to 255)

  In addition, the box at the far right displays the color of the pixel.

  Zooming in/out:
  To Zoom, select the amount of zoom you want from the zoom menu.
  Less than 100% zooms out and more than 100% zooms in. 
  The 100% zoom level will always return you to your orginal picture.
  
  filename: a string representing the location and name of picture.
  If no filename is given, a file-chooser opens."""

  if not filename:
    filename = _easygui.fileopenbox("Select an image", 
                                    _sys.argv[0], '*', 
                                    [ [ "*.jpg", "*.png", "*.bmp", "*.gif",
                                        "Image files" ] ])
    if not filename: 
      raise RuntimeError("No image file selected.")
  img = load_picture(filename)
  tool = PictureTool(img)
  tool.run_tool()

# --------------------------------------------------------------------
  
if __name__ == "__main__":
  if len(_sys.argv > 1):
    picture_tool(_sys.argv[1])
  else:
    picture_tool()

# --------------------------------------------------------------------
